@@ -5,8 +5,8 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 0.8
-;; Package-Requires: ((emacs "27.1"))
+;; Version: 0.9
+;; Package-Requires: ((emacs "27.1") (compat "29.1.4.0"))
 ;; Homepage: https://github.com/minad/org-modern
 
 ;; This file is part of GNU Emacs.
@@ -32,6 +32,7 @@
 
 ;;; Code:
 
+(require 'compat)
 (require 'org)
 (eval-when-compile
   (require 'cl-lib)
@@ -86,7 +87,7 @@ and TIME must be surrounded with space.  For the syntax, refer to
 (defcustom org-modern-table-vertical 3
   "Width of vertical table lines in pixels.
 Set to nil to hide the vertical lines."
-  :type '(choice (const nil) integer))
+  :type '(choice (const nil) natnum))
 
 (defcustom org-modern-table-horizontal 0.1
   "Prettify horizontal table lines."
@@ -193,7 +194,7 @@ all other blocks."
 This variable can also be set to an integer between 0 and 16,
 which specifies the offset of the block border from the edge of
 the window."
-  :type '(choice boolean integer))
+  :type '(choice boolean natnum))
 
 (defcustom org-modern-keyword t
   "Prettify keywords like #+title.
@@ -332,6 +333,7 @@ the font.")
 (defvar-local org-modern--font-lock-keywords nil)
 (defvar-local org-modern--star-cache nil)
 (defvar-local org-modern--hide-stars-cache nil)
+(defvar-local org-modern--reset-hide-leading-stars nil)
 (defvar-local org-modern--checkbox-cache nil)
 (defvar-local org-modern--progress-cache nil)
 (defvar-local org-modern--table-sp-width 0)
@@ -480,9 +482,11 @@ the font.")
                                   (nth (logand i 1)
                                        org-modern--hide-stars-cache))))
     (when org-modern-star
-      (put-text-property end (1+ end) 'display
-                         (aref org-modern--star-cache
-                               (min (1- (length org-modern--star-cache)) level))))))
+      (put-text-property
+       (if (eq org-modern-hide-stars 'leading) beg end)
+       (1+ end) 'display
+       (aref org-modern--star-cache
+             (min (1- (length org-modern--star-cache)) level))))))
 
 (defun org-modern--table ()
   "Prettify vertical table lines."
@@ -657,11 +661,9 @@ the font.")
         (0 (org-modern--checkbox)))))
    (when (or org-modern-star org-modern-hide-stars)
      `(("^\\(\\**\\)\\* "
-        ,@(and (not (eq org-modern-hide-stars t))
-               (or org-modern-star (stringp org-modern-hide-stars))
-               '((0 (org-modern--star))))
-        ,@(and (eq org-modern-hide-stars 'leading) '((1 '(face nil invisible org-modern))))
-        ,@(and (eq org-modern-hide-stars t) '((0 '(face nil invisible org-modern)))))))
+        (0 ,(if (eq org-modern-hide-stars t)
+                ''(face nil invisible org-modern)
+              '(org-modern--star))))))
    (when org-modern-horizontal-rule
      `(("^[ \t]*-\\{5,\\}$" 0
         '(face org-modern-horizontal-rule display
@@ -749,9 +751,15 @@ the font.")
   "Modern looks for Org."
   :global nil
   :group 'org-modern
+  (when org-modern--reset-hide-leading-stars
+    (setq-local org-hide-leading-stars t
+                org-modern--reset-hide-leading-stars nil))
   (cond
    (org-modern-mode
     (add-to-invisibility-spec 'org-modern)
+    (when (and (eq org-modern-hide-stars 'leading) org-hide-leading-stars)
+      (setq-local org-modern--reset-hide-leading-stars t
+                  org-hide-leading-stars nil))
     (setq
      org-modern--star-cache
      (vconcat (mapcar #'org-modern--symbol org-modern-star))
@@ -773,6 +781,8 @@ the font.")
     (font-lock-add-keywords nil org-modern--font-lock-keywords)
     (setq-local font-lock-unfontify-region-function #'org-modern--unfontify)
     (add-hook 'pre-redisplay-functions #'org-modern--pre-redisplay nil 'local)
+    (add-hook 'org-after-promote-entry-hook #'org-modern--unfontify-line nil 'local)
+    (add-hook 'org-after-demote-entry-hook #'org-modern--unfontify-line nil 'local)
     (org-modern--update-label-face)
     (org-modern--update-fringe-bitmaps))
    (t
@@ -780,13 +790,19 @@ the font.")
     (font-lock-remove-keywords nil org-modern--font-lock-keywords)
     (font-lock-add-keywords nil org-font-lock-keywords)
     (setq-local font-lock-unfontify-region-function #'org-unfontify-region)
-    (remove-hook 'pre-redisplay-functions #'org-modern--pre-redisplay 'local)))
+    (remove-hook 'pre-redisplay-functions #'org-modern--pre-redisplay 'local)
+    (remove-hook 'org-after-promote-entry-hook #'org-modern--unfontify-line 'local)
+    (remove-hook 'org-after-demote-entry-hook #'org-modern--unfontify-line 'local)))
   (save-restriction
     (widen)
     (with-silent-modifications
       (org-modern--unfontify (point-min) (point-max)))
     (font-lock-flush)))
 
+(defun org-modern--unfontify-line ()
+  "Unfontify prettified elements on current line."
+  (org-modern--unfontify (pos-bol) (pos-eol)))
+
 (defun org-modern--unfontify (beg end &optional _loud)
   "Unfontify prettified elements between BEG and END."
   (let ((font-lock-extra-managed-props
